/**
 * @file include/bin2llvmir/providers/asm_instruction.h
 * @brief Mapping of LLVM instructions to underlying ASM instructions.
 * @copyright (c) 2017 Avast Software, licensed under the MIT license
 */

#ifndef BIN2LLVMIR_PROVIDERS_ASM_INSTRUCTION_H
#define BIN2LLVMIR_PROVIDERS_ASM_INSTRUCTION_H

#include <capstone/capstone.h>
#include <llvm/IR/Instructions.h>
#include <llvm/IR/Module.h>

#include "llvm-support/utils.h"
#include "tl-cpputils/address.h"
#include "tl-cpputils/value.h"

namespace bin2llvmir {

class Config;

class AsmInstruction
{
	public:
		template<
			typename Category,
			typename Type,
			typename Reference = Type&,
			typename Pointer = Type*,
			typename Distance = std::ptrdiff_t>
		class iterator_impl
		{
			public:
				using difference_type = Distance;
				using value_type = Type;
				using reference = Reference;
				using pointer = Pointer;
				using iterator_category = Category;

			public:
				iterator_impl(llvm::StoreInst* s, bool end = false)
				{
					_first = s;
					_last = s;
					if (s == nullptr)
					{
						return;
					}
					llvm::Instruction* i = s;

					auto* bb = i->getParent();
					while (i && (i == _first
							|| !isLlvmToAsmInstruction(i)))
					{
						if (i != _first)
						{
							_last = i;
							if (_current == nullptr)
							{
								_current = i;
							}
						}

						if (&bb->back() == i)
						{
							if (&bb->getParent()->back() == bb)
							{
								break;
							}
							else
							{
								bb = bb->getNextNode();
								i = &bb->front();
							}
						}
						else
						{
							i = i->getNextNode();
						}
					}

					if (end)
					{
						_current = nullptr;
					}
				}
				iterator_impl() = default;
				iterator_impl(const iterator_impl& itr) = default;
				iterator_impl& operator=(iterator_impl rhs)
				{
					_first = rhs._first;
					_last = rhs._last;
					_current = rhs._current;
					return *this;
				}

				iterator_impl& operator++()
				{
					if (_current == nullptr)
					{
						return *this;
					}

					auto* bb = _current->getParent()->getNextNode();
					_current = _current->getNextNode();
					if (_current == nullptr)
					{
						if (bb)
						{
							_current = &bb->front();
						}
					}
					if (isLlvmToAsmInstruction(_current))
					{
						_current = nullptr;
					}
					return *this;
				}

				iterator_impl operator++(int)
				{
					if (_current == nullptr)
					{
						return *this;
					}

					iterator_impl tmp(*this);
					auto* bb = _current->getParent()->getNextNode();
					_current = _current->getNextNode();
					if (_current == nullptr)
					{
						if (bb)
						{
							_current = &bb->front();
						}
					}
					if (isLlvmToAsmInstruction(_current))
					{
						_current = nullptr;
					}
					return tmp;
				}

				iterator_impl& operator--()
				{
					if (_current == _first)
					{
						return *this;
					}
					if (_current == nullptr)
					{
						_current = _last;
						return *this;
					}

					auto* bb = _current->getParent()->getPrevNode();
					_current = _current->getPrevNode();
					if (_current == nullptr)
					{
						if (bb)
						{
							_current = &bb->back();
						}
					}
					return *this;
				}

				iterator_impl operator--(int)
				{
					if (_current == _first)
					{
						return *this;
					}

					iterator_impl tmp(*this);

					if (_current == nullptr)
					{
						_current = _last;
						return tmp;
					}

					auto* bb = _current->getParent()->getPrevNode();
					_current = _current->getPrevNode();
					if (_current == nullptr)
					{
						if (bb)
						{
							_current = &bb->back();
						}
					}
					return tmp;
				}

				reference operator*()
				{
					assert(_first != _current);
					return *_current;
				}

				pointer operator->()
				{
					assert(_first != _current);
					return &(*_current);
				}

				bool operator==(const iterator_impl& rhs) const
				{
					return (_current == nullptr && rhs._current == nullptr)
							|| (_first == rhs._first
									&& _last == rhs._last
									&& _current == rhs._current);
				}

				bool operator!=(const iterator_impl& rhs) const
				{
					return !(*this == rhs);
				}

			private:
				bool isLlvmToAsmInstruction(const llvm::Instruction* i) const
				{
					auto* s = llvm::dyn_cast_or_null<llvm::StoreInst>(i);
					return s &&
						s->getPointerOperand() == _first->getPointerOperand();
				}

			private:
				llvm::StoreInst* _first = nullptr;
				llvm::Instruction* _last = nullptr;
				llvm::Instruction* _current = nullptr;
		};

	public:
		using iterator = iterator_impl<
				std::bidirectional_iterator_tag,
				llvm::Instruction>;
		using const_iterator = iterator_impl<
				std::bidirectional_iterator_tag,
				const llvm::Instruction>;
		using reverse_iterator = std::reverse_iterator<iterator>;
		using const_reverse_iterator = std::reverse_iterator<const_iterator>;

		iterator begin();
		iterator end();
		reverse_iterator rbegin();
		reverse_iterator rend();
		const_iterator begin() const;
		const_iterator end() const;
		const_reverse_iterator rbegin() const;
		const_reverse_iterator rend() const;

	public:
		AsmInstruction();
		AsmInstruction(llvm::Instruction* inst);
		AsmInstruction(llvm::BasicBlock* bb);
		AsmInstruction(llvm::Function* f);
		AsmInstruction(llvm::Module* m, tl_cpputils::Address addr);

		bool operator<(const AsmInstruction& o) const;
		bool operator==(const AsmInstruction& o) const;
		bool operator!=(const AsmInstruction& o) const;
		explicit operator bool() const;

		bool isValid() const;
		bool isInvalid() const;
		bool isConditional(Config* conf) const;
		cs_insn* getCapstoneInsn() const;
		bool isThumb() const;

		std::string getDsm() const;
		tl_cpputils::Maybe<unsigned> getLatency() const;
		tl_cpputils::Address getAddress() const;
		tl_cpputils::Address getEndAddress() const;
		std::size_t getByteSize() const;
		std::size_t getBitSize() const;
		bool contains(tl_cpputils::Address addr) const;

		AsmInstruction getNext() const;
		AsmInstruction getPrev() const;

		bool instructionsCanBeErased();
		bool eraseInstructions();
		llvm::TerminatorInst* makeTerminal();
		llvm::BasicBlock* makeStart();

		llvm::BasicBlock* getBasicBlock() const;
		llvm::Function* getFunction() const;
		std::vector<llvm::Instruction*> getInstructions();
		std::vector<llvm::BasicBlock*> getBasicBlocks();
		std::string getBasicBlockLableName(
				const std::string& labelPrefix = "dec_label_pc_") const;

		bool empty();
		llvm::Instruction* front();
		llvm::Instruction* back();
		llvm::StoreInst* getLlvmToAsmInstruction() const;

		llvm::Instruction* insertBack(llvm::Instruction* i);
		llvm::Instruction* insertBackSafe(llvm::Instruction* i);

		bool storesValue(llvm::Value* val) const;

		std::string dump() const;
		friend std::ostream& operator<<(
				std::ostream& out,
				const AsmInstruction &a);

	// Templates.
	//
	public:
		template<typename T>
		bool containsInstruction()
		{
			for (auto& i : *this)
			{
				if (llvm::isa<T>(&i))
				{
					return true;
				}
			}
			return false;
		}

		template<typename T>
		T* getInstructionFirst()
		{
			for (auto& i : *this)
			{
				if (auto* ret = llvm::dyn_cast<T>(&i))
				{
					return ret;
				}
			}
			return nullptr;
		}

	public:
		static const llvm::GlobalVariable* getLlvmToAsmGlobalVariable(
				const llvm::Module* m);
		static tl_cpputils::Address getInstructionAddress(
				llvm::Instruction* inst);
		static bool isLlvmToAsmInstruction(const llvm::Value* inst);
		static void clear();

	private:
		const llvm::GlobalVariable* getLlvmToAsmGlobalVariablePrivate(
				llvm::Module* m) const;
		bool isLlvmToAsmInstructionPrivate(llvm::Value* inst) const;

	private:
		using ModuleGlobalPair = std::pair<const llvm::Module*, const llvm::GlobalVariable*>;

	private:
		llvm::StoreInst* _llvmToAsmInstr = nullptr;
		static std::vector<ModuleGlobalPair> _cache;
};

} // namespace bin2llvmir

#endif
